import os
import subprocess
import uuid


def jinja_global_cmd(cmd: str) -> str:
    """
    Runs a system command provided as argument and returns the output

    Args:
        cmd (str): a command to execute
    """
    result = subprocess.run(
        cmd,
        shell=True,
        check=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True
    )
    return result.stdout


def jinja_global_cwd() -> str:
    """
    Returns the current working directory
    """
    return os.getcwd()


def jinja_global_env(name: str, default: str | None = "") -> str | None:
    """
    Returns value of an environment variable

    Args:
        name (str): name of environment variable
        default (str | None): a default value to return
    """
    v = os.getenv(name, default)
    if v is None:
        raise ValueError(f"An environment variable '{name}' is not set and no default value is provided.")
    return v


def jinja_global_file(path: str) -> str:
    """
    Returns contents of file

    Args:
        path (str): path to file
    """
    with open(path, "r") as f:
        return f.read()


def jinja_global_uuid() -> str:
    """
    Generates UUID
    """
    return str(uuid.uuid4())



import base64
import hashlib


def jinja_filter_split_space(input: str) -> list[str]:
    """Splits a multiline string into list of strings where each item is a line"""
    return input.strip().split()


def jinja_filter_md5(input_string: str) -> str:
    return hashlib.md5(input_string.encode()).hexdigest()


def jinja_filter_sha256(input_string: str) -> str:
    return hashlib.sha256(input_string.encode()).hexdigest()


def jinja_filter_sha512(input_string: str) -> str:
    return hashlib.sha512(input_string.encode()).hexdigest()


def jinja_filter_base64(input_string: str) -> str:
    return base64.b64encode(input_string.encode()).decode()


def jinja_filter_base64_decode(encoded_string: str) -> str:
    return base64.b64decode(encoded_string.encode()).decode()



from copy import deepcopy
import os
import os.path
from typing import Callable
import yaml

from configtpl.utils.dicts import dict_deep_merge
from configtpl.jinja.env_factory import JinjaEnvFactory


class ConfigBuilder:
    def __init__(self, jinja_constructor_args: dict | None = None, jinja_globals: dict | None = None,
                 jinja_filters: dict | None = None, defaults: dict | None = None):
        """
        A constructor for Cofnig Builder.

        Args:
            constructor_args (dict | None): argument for Jinja environment constructor
            globals (dict | None): globals to inject into Jinja environment
            defaults (dict | None): Default values for configuration
        """
        self.jinja_env_factory = JinjaEnvFactory(constructor_args=jinja_constructor_args, globals=jinja_globals,
                                                 filters=jinja_filters)
        if defaults is None:
            defaults = {}
        self.defaults = defaults

    def set_global(self, k: str, v: Callable) -> None:
        """
        Sets a global for children Jinja environments
        """
        self.jinja_env_factory.set_global(k, v)

    def set_filter(self, k: str, v: Callable) -> None:
        """
        Sets a filter for children Jinja environments
        """
        self.jinja_env_factory.set_filter(k, v)

    def build_from_files(self, paths_colon_separated: str | list[str], overrides: dict | None = None,
                         ctx: dict | None = None) -> dict:
        """
        Renders files from provided paths.

        Args:
            ctx (dict | None): additional rendering context which is NOT injected into configuration
            overrides (dict | None): Overrides are applied at the very end stage after all templates are rendered
            paths_colon_separated (str | list[str]): Paths to configuration files.
                It might be a single item (str) or list of paths (list(str)).
                Additionally, each path might be colon-separated.
                Examples: '/opt/myapp/myconfig.cfg', '/opt/myapp/myconfig_first.cfg:/opt/myapp/myconfig_second.cfg',
                ['/opt/myapp/myconfig.cfg', '/opt/myapp/myconfig_first.cfg:/opt/myapp/myconfig_second.cfg']
        Returns:
            dict: The rendered configuration
        """
        output_cfg = deepcopy(self.defaults)
        if ctx is None:
            ctx = {}
        if overrides is None:
            overrides = {}
        # Convert the path input into list of paths
        paths_colon_separated: list[str] = [paths_colon_separated] \
            if isinstance(paths_colon_separated, str) \
            else paths_colon_separated
        paths: list[str] = []
        for path_colon_separated in paths_colon_separated:
            paths += path_colon_separated.split(":")

        for cfg_path in paths:
            cfg_path = os.path.realpath(cfg_path)
            ctx = {**output_cfg, **ctx}
            cfg_iter: dict = self._render_cfg_from_file(cfg_path, ctx)

            output_cfg = dict_deep_merge(output_cfg, cfg_iter)

        # Append overrides
        output_cfg = dict_deep_merge(output_cfg, overrides)
        return output_cfg

    def build_from_str(self, input: str, work_dir: str | None = None, defaults: dict | None = None,
                       ctx: dict | None = None, overrides: dict | None = None) -> dict:
        """
        Renders config from string.

        Args:
            input (str): a Jinja template string which can be rendered into YAML format
            work_dir (str): a working directory.
                Include statements in Jinja template will be resolved relatively to this path
            defaults (dict | None): Default values for configuration
            ctx (dict | None): additional rendering context which is NOT injected into configuration
            overrides (dict | None): Overrides are applied at the very end stage after all templates are rendered
        Returns:
            dict: The rendered configuration
        """
        if work_dir is None:
            work_dir = os.getcwd()
        output_cfg = {} if defaults is None else deepcopy(defaults)
        if ctx is None:
            ctx = {}
        if overrides is None:
            overrides = {}

        cfg = self._render_cfg_from_str(input, ctx, work_dir)
        output_cfg = dict_deep_merge(cfg, overrides)
        return output_cfg

    def _render_cfg_from_file(self, path: str, ctx: dict) -> dict:
        """
        Renders a template file into config dictionary in two steps:
        1. Renders a file as Jinja template
        2. Parses the rendered file as YAML template
        """
        dir = os.path.dirname(path)
        filename = os.path.basename(path)
        jinja_env = self.jinja_env_factory.get_fs_jinja_environment(dir)
        tpl = jinja_env.get_template(filename)
        tpl_rendered = tpl.render(ctx)
        return yaml.load(tpl_rendered, Loader=yaml.FullLoader)

    def _render_cfg_from_str(self, input: str, ctx: dict, work_dir: str) -> dict:
        jinja_env = self.jinja_env_factory.get_fs_jinja_environment(work_dir)
        tpl = jinja_env.from_string(input)
        tpl_rendered = tpl.render(ctx)
        return yaml.load(tpl_rendered, Loader=yaml.FullLoader)



import jinja2
from typing import Callable

from configtpl.jinja import filters as jinja_filters
from configtpl.jinja import globals as jinja_globals
from configtpl.utils.dicts import dict_deep_merge


class JinjaEnvFactory:
    def __init__(self, constructor_args: dict | None = None, globals: dict | None = None, filters: dict | None = None):
        """
        A constructor for Jinja Envoronment Factory

        Args:
            constructor_args (dict | None): argument for Jinja environment constructor
            globals (dict | None): globals to inject into Jinja environment
        """
        self._constructor_args = dict_deep_merge(
            {
                "undefined": jinja2.StrictUndefined,
            },
            {} if constructor_args is None else constructor_args,
        )
        self._globals = dict_deep_merge(
            {
                "cmd": jinja_globals.jinja_global_cmd,
                "cwd": jinja_globals.jinja_global_cwd,
                "env": jinja_globals.jinja_global_env,
                "file": jinja_globals.jinja_global_file,
                "uuid": jinja_globals.jinja_global_uuid,
            },
            {} if globals is None else globals,
        )
        self._filters = dict_deep_merge(
            {
                "base64": jinja_filters.jinja_filter_base64,
                "base64_decode": jinja_filters.jinja_filter_base64_decode,
                "md5": jinja_filters.jinja_filter_md5,
                "split_space": jinja_filters.jinja_filter_split_space,
                "sha256": jinja_filters.jinja_filter_sha256,
                "sha512": jinja_filters.jinja_filter_sha512,
            },
            {} if filters is None else filters,
        )

        self._fs_loader_cache = {}

    def set_global(self, k: str, v: Callable) -> None:
        """
        Sets a global for children Jinja environments
        """
        self._globals[k] = v

    def set_filter(self, k: str, v: Callable) -> None:
        """
        Sets a filter for children Jinja environments
        """
        self._filters[k] = v

    def get_fs_jinja_environment(self, dir: str) -> jinja2.Environment:
        """
        Creates an instance of Jinja environment with filesystem loaded for provided directory
        """
        if dir in self._fs_loader_cache:
            return self._fs_loader_cache[dir]

        jinja_env = jinja2.Environment(**self._constructor_args, loader=jinja2.FileSystemLoader(dir))
        jinja_env.globals.update(self._globals)
        jinja_env.filters.update(self._filters)

        self._fs_loader_cache[dir] = jinja_env

        return jinja_env



def dict_deep_merge(*dicts: dict) -> dict:
    """
    Deep merge multiple dictionaries recursively.
    Values in later dictionaries overwrite those in earlier ones.
    This function does not update any dictionary by reference.
    """
    def merge_two_dicts(d1: dict, d2: dict):
        merged = d1.copy()
        for key, value in d2.items():
            if key in merged and isinstance(merged[key], dict) and isinstance(value, dict):
                merged[key] = merge_two_dicts(merged[key], value)
            else:
                merged[key] = value
        return merged

    result = {}
    for d in dicts:
        result = merge_two_dicts(result, d)

    return result
